"""Deprecated as of 5.0; import from IPython.lib.lexers instead."""

from warnings import warn

warn("nbconvert.utils.lexers is deprecated as of 5.0. Use IPython.lib.lexers", stacklevel=2)

from IPython.lib.lexers import *  # noqa: F403, E402



"""Contains writer for writing nbconvert output to filesystem."""

# Copyright (c) IPython Development Team.
# Distributed under the terms of the Modified BSD License.

import errno
import glob
import os
from pathlib import Path

from traitlets import Unicode, observe

from nbconvert.utils.io import link_or_copy

from .base import WriterBase


class FilesWriter(WriterBase):
    """Consumes nbconvert output and produces files."""

    build_directory = Unicode(
        "",
        help="""Directory to write output(s) to. Defaults
                              to output to the directory of each notebook. To recover
                              previous default behaviour (outputting to the current
                              working directory) use . as the flag value.""",
    ).tag(config=True)

    relpath = Unicode(
        help="""When copying files that the notebook depends on, copy them in
        relation to this path, such that the destination filename will be
        os.path.relpath(filename, relpath). If FilesWriter is operating on a
        notebook that already exists elsewhere on disk, then the default will be
        the directory containing that notebook."""
    ).tag(config=True)

    # Make sure that the output directory exists.
    @observe("build_directory")
    def _build_directory_changed(self, change):
        new = change["new"]
        if new:
            self._makedir(new)

    def __init__(self, **kw):
        """Initialize the writer."""
        super().__init__(**kw)
        self._build_directory_changed({"new": self.build_directory})

    def _makedir(self, path, mode=0o755):
        """ensure that a directory exists

        If it doesn't exist, try to create it and protect against a race condition
        if another process is doing the same.

        The default permissions are 755, which differ from os.makedirs default of 777.
        """
        if not os.path.exists(path):
            self.log.info("Making directory %s", path)
            try:
                os.makedirs(path, mode=mode)
            except OSError as e:
                if e.errno != errno.EEXIST:
                    raise
        elif not os.path.isdir(path):
            raise OSError("%r exists but is not a directory" % path)

    def _write_items(self, items, build_dir):
        """Write a dict containing filename->binary data"""
        for filename, data in items:
            # Determine where to write the file to
            dest = os.path.join(build_dir, filename)
            path = os.path.dirname(dest)
            self._makedir(path)

            # Write file
            self.log.debug("Writing %i bytes to %s", len(data), dest)
            with open(dest, "wb") as f:
                f.write(data)

    def write(self, output, resources, notebook_name=None, **kw):
        """
        Consume and write Jinja output to the file system.  Output directory
        is set via the 'build_directory' variable of this instance (a
        configurable).

        See base for more...
        """

        # Verify that a notebook name is provided.
        if notebook_name is None:
            msg = "notebook_name"
            raise TypeError(msg)

        # Pull the extension and subdir from the resources dict.
        output_extension = resources.get("output_extension", None)

        # Get the relative path for copying files
        resource_path = resources.get("metadata", {}).get("path", "")
        relpath = self.relpath or resource_path
        build_directory = self.build_directory or resource_path

        # Write the extracted outputs to the destination directory.
        # NOTE: WE WRITE EVERYTHING AS-IF IT'S BINARY.  THE EXTRACT FIG
        # PREPROCESSOR SHOULD HANDLE UNIX/WINDOWS LINE ENDINGS...

        items = resources.get("outputs", {}).items()
        if items:
            self.log.info(
                "Support files will be in %s",
                os.path.join(resources.get("output_files_dir", ""), ""),
            )
            self._write_items(items, build_directory)

        # Write the extracted attachments
        # if ExtractAttachmentsOutput specified a separate directory
        attachments = resources.get("attachments", {}).items()
        if attachments:
            self.log.info(
                "Attachments will be in %s",
                os.path.join(resources.get("attachment_files_dir", ""), ""),
            )
            self._write_items(attachments, build_directory)

        # Copy referenced files to output directory
        if build_directory:
            for filename in self.files:
                # Copy files that match search pattern
                for matching_filename in glob.glob(filename):
                    # compute the relative path for the filename
                    if relpath != "":
                        dest_filename = os.path.relpath(matching_filename, relpath)
                    else:
                        dest_filename = matching_filename

                    # Make sure folder exists.
                    dest = os.path.join(build_directory, dest_filename)
                    path = os.path.dirname(dest)
                    self._makedir(path)

                    # Copy if destination is different.
                    if os.path.normpath(dest) != os.path.normpath(matching_filename):
                        self.log.info("Copying %s -> %s", matching_filename, dest)
                        link_or_copy(matching_filename, dest)

        # Determine where to write conversion results.
        dest = notebook_name + output_extension if output_extension is not None else notebook_name
        dest_path = Path(build_directory) / dest

        # Write conversion results.
        self.log.info("Writing %i bytes to %s", len(output), dest_path)
        if isinstance(output, str):
            with open(dest_path, "w", encoding="utf-8") as f:
                f.write(output)
        else:
            with open(dest_path, "wb") as f:
                f.write(output)

        return dest_path



"""NotebookExporter class"""

# Copyright (c) IPython Development Team.
# Distributed under the terms of the Modified BSD License.

import nbformat
from traitlets import Enum, default

from .exporter import Exporter


class NotebookExporter(Exporter):
    """Exports to an IPython notebook.

    This is useful when you want to use nbconvert's preprocessors to operate on
    a notebook (e.g. to execute it) and then write it back to a notebook file.
    """

    nbformat_version = Enum(
        list(nbformat.versions),
        default_value=nbformat.current_nbformat,
        help="""The nbformat version to write.
        Use this to downgrade notebooks.
        """,
    ).tag(config=True)

    @default("file_extension")
    def _file_extension_default(self):
        return ".ipynb"

    output_mimetype = "application/json"
    export_from_notebook = "Notebook"

    def from_notebook_node(self, nb, resources=None, **kw):
        """Convert from notebook node."""
        nb_copy, resources = super().from_notebook_node(nb, resources, **kw)
        if self.nbformat_version != nb_copy.nbformat:
            resources["output_suffix"] = ".v%i" % self.nbformat_version
        else:
            resources["output_suffix"] = ".nbconvert"
        output = nbformat.writes(nb_copy, version=self.nbformat_version)
        if not output.endswith("\n"):
            output = output + "\n"
        return output, resources



"""HTML slide show Exporter class"""

# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.

from copy import deepcopy
from warnings import warn

from traitlets import Bool, Unicode, default

from nbconvert.preprocessors.base import Preprocessor

from .html import HTMLExporter


class _RevealMetadataPreprocessor(Preprocessor):
    # A custom preprocessor adding convenience metadata to cells

    def preprocess(self, nb, resources=None):
        nb = deepcopy(nb)

        for cell in nb.cells:
            # Make sure every cell has a slide_type
            try:
                slide_type = cell.metadata.get("slideshow", {}).get("slide_type", "-")
            except AttributeError:
                slide_type = "-"
            cell.metadata.slide_type = slide_type

        # Find the first visible cell
        for index, cell in enumerate(nb.cells):
            if cell.metadata.slide_type not in {"notes", "skip"}:
                cell.metadata.slide_type = "slide"
                cell.metadata.slide_start = True
                cell.metadata.subslide_start = True
                first_slide_ix = index
                break
        else:
            msg = "All cells are hidden, cannot create slideshow"
            raise ValueError(msg)

        in_fragment = False

        for index, cell in enumerate(nb.cells[first_slide_ix + 1 :], start=(first_slide_ix + 1)):
            previous_cell = nb.cells[index - 1]

            # Slides are <section> elements in the HTML, subslides (the vertically
            # stacked slides) are also <section> elements inside the slides,
            # and fragments are <div>s within subslides. Subslide and fragment
            # elements can contain content:
            # <section>
            #   <section>
            #     (content)
            #     <div class="fragment">(content)</div>
            #   </section>
            # </section>

            # Get the slide type. If type is subslide or slide,
            # end the last slide/subslide/fragment as applicable.
            if cell.metadata.slide_type == "slide":
                previous_cell.metadata.slide_end = True
                cell.metadata.slide_start = True
            if cell.metadata.slide_type in {"subslide", "slide"}:
                previous_cell.metadata.fragment_end = in_fragment
                previous_cell.metadata.subslide_end = True
                cell.metadata.subslide_start = True
                in_fragment = False

            elif cell.metadata.slide_type == "fragment":
                cell.metadata.fragment_start = True
                if in_fragment:
                    previous_cell.metadata.fragment_end = True
                else:
                    in_fragment = True

        # The last cell will always be the end of a slide
        nb.cells[-1].metadata.fragment_end = in_fragment
        nb.cells[-1].metadata.subslide_end = True
        nb.cells[-1].metadata.slide_end = True

        return nb, resources


class SlidesExporter(HTMLExporter):
    """Exports HTML slides with reveal.js"""

    # Overrides from HTMLExporter
    #################################
    export_from_notebook = "Reveal.js slides"

    @default("template_name")
    def _template_name_default(self):
        return "reveal"

    @default("file_extension")
    def _file_extension_default(self):
        return ".slides.html"

    @default("template_extension")
    def _template_extension_default(self):
        return ".html.j2"

    # Extra resources
    #################################
    reveal_url_prefix = Unicode(
        help="""The URL prefix for reveal.js (version 3.x).
        This defaults to the reveal CDN, but can be any url pointing to a copy
        of reveal.js.

        For speaker notes to work, this must be a relative path to a local
        copy of reveal.js: e.g., "reveal.js".

        If a relative path is given, it must be a subdirectory of the
        current directory (from which the server is run).

        See the usage documentation
        (https://nbconvert.readthedocs.io/en/latest/usage.html#reveal-js-html-slideshow)
        for more details.
        """
    ).tag(config=True)

    @default("reveal_url_prefix")
    def _reveal_url_prefix_default(self):
        if "RevealHelpPreprocessor.url_prefix" in self.config:
            warn(
                "Please update RevealHelpPreprocessor.url_prefix to "
                "SlidesExporter.reveal_url_prefix in config files.",
                stacklevel=2,
            )
            return self.config.RevealHelpPreprocessor.url_prefix
        return "https://unpkg.com/reveal.js@4.0.2"

    reveal_theme = Unicode(
        "simple",
        help="""
        Name of the reveal.js theme to use.

        We look for a file with this name under
        ``reveal_url_prefix``/css/theme/``reveal_theme``.css.

        https://github.com/hakimel/reveal.js/tree/master/css/theme has
        list of themes that ship by default with reveal.js.
        """,
    ).tag(config=True)

    reveal_transition = Unicode(
        "slide",
        help="""
        Name of the reveal.js transition to use.

        The list of transitions that ships by default with reveal.js are:
        none, fade, slide, convex, concave and zoom.
        """,
    ).tag(config=True)

    reveal_scroll = Bool(
        False,
        help="""
        If True, enable scrolling within each slide
        """,
    ).tag(config=True)

    reveal_number = Unicode(
        "",
        help="""
        slide number format (e.g. 'c/t'). Choose from:
        'c': current, 't': total, 'h': horizontal, 'v': vertical
        """,
    ).tag(config=True)

    reveal_width = Unicode(
        "",
        help="""
        width used to determine the aspect ratio of your presentation.
        Use the horizontal pixels available on your intended presentation
        equipment.
        """,
    ).tag(config=True)

    reveal_height = Unicode(
        "",
        help="""
        height used to determine the aspect ratio of your presentation.
        Use the horizontal pixels available on your intended presentation
        equipment.
        """,
    ).tag(config=True)

    font_awesome_url = Unicode(
        "https://cdnjs.cloudflare.com/ajax/libs/font-awesome/4.7.0/css/font-awesome.css",
        help="""
        URL to load font awesome from.

        Defaults to loading from cdnjs.
        """,
    ).tag(config=True)

    def _init_resources(self, resources):
        resources = super()._init_resources(resources)
        if "reveal" not in resources:
            resources["reveal"] = {}
        resources["reveal"]["url_prefix"] = self.reveal_url_prefix
        resources["reveal"]["theme"] = self.reveal_theme
        resources["reveal"]["transition"] = self.reveal_transition
        resources["reveal"]["scroll"] = self.reveal_scroll
        resources["reveal"]["number"] = self.reveal_number
        resources["reveal"]["height"] = self.reveal_height
        resources["reveal"]["width"] = self.reveal_width
        return resources



"""
Module that extracts attachments from notebooks into their own files
"""

# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.

import os
from base64 import b64decode

from traitlets import Bool, Unicode

from .base import Preprocessor


class ExtractAttachmentsPreprocessor(Preprocessor):
    """
    Extracts attachments from all (markdown and raw) cells in a notebook.
    The extracted attachments are stored in a directory ('attachments' by default).
    https://nbformat.readthedocs.io/en/latest/format_description.html#cell-attachments
    """

    attachments_directory_template = Unicode(
        "{notebook_name}_attachments",
        help="Directory to place attachments if use_separate_dir is True",
    ).tag(config=True)

    use_separate_dir = Bool(
        False,
        help="Whether to use output_files_dir (which ExtractOutput also uses) or "
        "create a separate directory for attachments",
    ).tag(config=True)

    def __init__(self, **kw):
        """
        Public constructor
        """
        super().__init__(**kw)
        # directory path,
        self.path_name = ""  # will be set in self.preprocess, needs resources
        # Where extracted attachments are stored in resources
        self.resources_item_key = (
            "attachments"  # Here as a default, in case someone doesn't want to call preprocess
        )

    # Add condition and configurability here
    def preprocess(self, nb, resources):
        """
        Determine some settings and apply preprocessor to notebook
        """
        if self.use_separate_dir:
            self.path_name = self.attachments_directory_template.format(
                notebook_name=resources["unique_key"]
            )
            # Initialize resources for attachments
            resources["attachment_files_dir"] = self.path_name
            resources["attachments"] = {}
            self.resources_item_key = "attachments"
        else:
            # Use same resources as ExtractOutput
            self.path_name = resources["output_files_dir"]
            self.resources_item_key = "outputs"

        # Make sure key exists
        if not isinstance(resources[self.resources_item_key], dict):
            resources[self.resources_item_key] = {}

        nb, resources = super().preprocess(nb, resources)
        return nb, resources

    def preprocess_cell(self, cell, resources, index):
        """
        Extract attachments to individual files and
        change references to them.
        E.g.
        '![image.png](attachment:021fdd80.png)'
        becomes
        '![image.png]({path_name}/021fdd80.png)'
        Assumes self.path_name and self.resources_item_key is set properly (usually in preprocess).
        """
        if "attachments" in cell:
            for fname in cell.attachments:
                self.log.debug("Encountered attachment %s", fname)

                # Add file for writer

                # Right now I don't know of a situation where there would be multiple
                # mime types under same filename, and I can't index into it without the mimetype.
                # So I only read the first one.
                for mimetype in cell.attachments[fname]:
                    # convert to bytes and decode
                    data = cell.attachments[fname][mimetype].encode("utf-8")
                    decoded = b64decode(data)
                    break

                # FilesWriter wants path to be in attachment filename here
                new_filename = os.path.join(self.path_name, fname)
                resources[self.resources_item_key][new_filename] = decoded

                # Edit the reference to the attachment

                # os.path.join on windows uses "\\" separator,
                # but files like markdown still want "/"
                if os.path.sep != "/":
                    new_filename = new_filename.replace(os.path.sep, "/")
                cell.source = cell.source.replace("attachment:" + fname, new_filename)

        return cell, resources



"""Module containing a preprocessor that removes the outputs from code cells"""

# Copyright (c) IPython Development Team.
# Distributed under the terms of the Modified BSD License.

from traitlets import Set

from .base import Preprocessor


class ClearOutputPreprocessor(Preprocessor):
    """
    Removes the output from all code cells in a notebook.
    """

    remove_metadata_fields = Set({"collapsed", "scrolled"}).tag(config=True)

    def preprocess_cell(self, cell, resources, cell_index):
        """
        Apply a transformation on each cell. See base.py for details.
        """
        if cell.cell_type == "code":
            cell.outputs = []
            cell.execution_count = None
            # Remove metadata associated with output
            if "metadata" in cell:
                for field in self.remove_metadata_fields:
                    cell.metadata.pop(field, None)
        return cell, resources



"""Export to PDF via a headless browser"""

# Copyright (c) IPython Development Team.
# Distributed under the terms of the Modified BSD License.

import asyncio
import concurrent.futures
import os
import subprocess
import sys
import tempfile
from importlib import util as importlib_util

from traitlets import Bool, default

from .html import HTMLExporter

PLAYWRIGHT_INSTALLED = importlib_util.find_spec("playwright") is not None
IS_WINDOWS = os.name == "nt"


class WebPDFExporter(HTMLExporter):
    """Writer designed to write to PDF files.

    This inherits from :class:`HTMLExporter`. It creates the HTML using the
    template machinery, and then run playwright to create a pdf.
    """

    export_from_notebook = "PDF via HTML"

    allow_chromium_download = Bool(
        False,
        help="Whether to allow downloading Chromium if no suitable version is found on the system.",
    ).tag(config=True)

    paginate = Bool(
        True,
        help="""
        Split generated notebook into multiple pages.

        If False, a PDF with one long page will be generated.

        Set to True to match behavior of LaTeX based PDF generator
        """,
    ).tag(config=True)

    @default("file_extension")
    def _file_extension_default(self):
        return ".html"

    @default("template_name")
    def _template_name_default(self):
        return "webpdf"

    disable_sandbox = Bool(
        False,
        help="""
        Disable chromium security sandbox when converting to PDF.

        WARNING: This could cause arbitrary code execution in specific circumstances,
        where JS in your notebook can execute serverside code! Please use with
        caution.

        ``https://github.com/puppeteer/puppeteer/blob/main@%7B2020-12-14T17:22:24Z%7D/docs/troubleshooting.md#setting-up-chrome-linux-sandbox``
        has more information.

        This is required for webpdf to work inside most container environments.
        """,
    ).tag(config=True)

    def run_playwright(self, html):
        """Run playwright."""

        async def main(temp_file):
            """Run main playwright script."""
            args = ["--no-sandbox"] if self.disable_sandbox else []
            try:
                from playwright.async_api import async_playwright  # type: ignore[import-not-found]
            except ModuleNotFoundError as e:
                msg = (
                    "Playwright is not installed to support Web PDF conversion. "
                    "Please install `nbconvert[webpdf]` to enable."
                )
                raise RuntimeError(msg) from e

            if self.allow_chromium_download:
                cmd = [sys.executable, "-m", "playwright", "install", "chromium"]
                subprocess.check_call(cmd)  # noqa: S603

            playwright = await async_playwright().start()
            chromium = playwright.chromium

            try:
                browser = await chromium.launch(
                    handle_sigint=False, handle_sigterm=False, handle_sighup=False, args=args
                )
            except Exception as e:
                msg = (
                    "No suitable chromium executable found on the system. "
                    "Please use '--allow-chromium-download' to allow downloading one,"
                    "or install it using `playwright install chromium`."
                )
                await playwright.stop()
                raise RuntimeError(msg) from e

            page = await browser.new_page()
            await page.emulate_media(media="print")
            await page.wait_for_timeout(100)
            await page.goto(f"file://{temp_file.name}", wait_until="networkidle")
            await page.wait_for_timeout(100)

            pdf_params = {"print_background": True}
            if not self.paginate:
                # Floating point precision errors cause the printed
                # PDF from spilling over a new page by a pixel fraction.
                dimensions = await page.evaluate(
                    """() => {
                    const rect = document.body.getBoundingClientRect();
                    return {
                    width: Math.ceil(rect.width) + 1,
                    height: Math.ceil(rect.height) + 1,
                    }
                }"""
                )
                width = dimensions["width"]
                height = dimensions["height"]
                # 200 inches is the maximum size for Adobe Acrobat Reader.
                pdf_params.update(
                    {
                        "width": min(width, 200 * 72),
                        "height": min(height, 200 * 72),
                    }
                )
            pdf_data = await page.pdf(**pdf_params)

            await browser.close()
            await playwright.stop()
            return pdf_data

        pool = concurrent.futures.ThreadPoolExecutor()
        # Create a temporary file to pass the HTML code to Chromium:
        # Unfortunately, tempfile on Windows does not allow for an already open
        # file to be opened by a separate process. So we must close it first
        # before calling Chromium. We also specify delete=False to ensure the
        # file is not deleted after closing (the default behavior).
        temp_file = tempfile.NamedTemporaryFile(suffix=".html", delete=False)
        with temp_file:
            temp_file.write(html.encode("utf-8"))
        try:
            # TODO: when dropping Python 3.6, use
            # pdf_data = pool.submit(asyncio.run, main(temp_file)).result()
            def run_coroutine(coro):
                """Run an internal coroutine."""
                loop = (
                    asyncio.ProactorEventLoop()  # type:ignore[attr-defined]
                    if IS_WINDOWS
                    else asyncio.new_event_loop()
                )

                asyncio.set_event_loop(loop)
                return loop.run_until_complete(coro)

            pdf_data = pool.submit(run_coroutine, main(temp_file)).result()
        finally:
            # Ensure the file is deleted even if playwright raises an exception
            os.unlink(temp_file.name)
        return pdf_data

    def from_notebook_node(self, nb, resources=None, **kw):
        """Convert from a notebook node."""
        html, resources = super().from_notebook_node(nb, resources=resources, **kw)

        self.log.info("Building PDF")
        pdf_data = self.run_playwright(html)
        self.log.info("PDF successfully created")

        # convert output extension to pdf
        # the writer above required it to be html
        resources["output_extension"] = ".pdf"

        return pdf_data, resources



"""Global configuration class."""

# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.

from traitlets import List, Unicode
from traitlets.config.configurable import LoggingConfigurable


class NbConvertBase(LoggingConfigurable):
    """Global configurable class for shared config

    Useful for display data priority that might be used by many transformers
    """

    display_data_priority = List(
        [
            "text/html",
            "application/pdf",
            "text/latex",
            "image/svg+xml",
            "image/png",
            "image/jpeg",
            "text/markdown",
            "text/plain",
        ],
        help="""
            An ordered list of preferred output type, the first
            encountered will usually be used when converting discarding
            the others.
            """,
    ).tag(config=True)

    default_language = Unicode(
        "ipython",
        help="Deprecated default highlight language as of 5.0, please use language_info metadata instead",
    ).tag(config=True)
